// Copyright (C) 1999 Dominic Letourneau and Andre Charbonneau
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2, or (at your option)
// any later version.
//
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this file.  If not, write to the Free Software Foundation,
// 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

#ifndef FileParser_h
#define FileParser_h
 

/*
** Includes
*/
#include <vector>
#include <list>
#include <string>
#include "FileParserException.h"
#include "FP_network_info_container.h"

/** Data-flow network file parser class.
 ** A FileParser class posseses methods to read and parse a data-flow network text file.
 ** The information needed to build the network is then stored in an indexed structure.
 ** For more information about the syntax of the data-flow network text file see the
 ** syntax documentation written by Andre Charbonneau.
  */
class FileParser
{
 private:
  /// The filename of the last data-flow network text file that was parsed.
  string m_filename;

  /// The indexed structure that will hold all the information needed to build a data-flow network.
  map<string,FP_network_info_container> m_networks_map;

  /// The verbose flag.  When this flag is set to true, then messages will be sent to cout as the file is being parsed.
  bool m_verbose;

  /* private methods */

  string::size_type get_start_index(const string& substring, const string& target, const string::size_type starting_at = 0) const;

  /** Read the content of a text file and stores it in a string object.
   ** @param filename The file to read.
   ** @return The content of the file as a string object.
   ** @exception FileParserException
   */
  string read_file(const string& filename) throw(FileParserException);


  /** Preprocesses a string (removes whitespaces and comments)
   ** @param s The string to preprocess.
   ** @return A string containing the preprocessed text.
   ** @exception FileParserException
   ** @see remove_comments
   ** @see remove_white_space
   */
  string preprocess(const string& s) const throw(FileParserException);


  /** Removes comments from a string.
   ** @param s The string to remove comments from.
   ** @return A string containing the same text as s but without comments.
   ** @exception FileParserException
   ** @memo This function will remove C style comments only.  Please consult the
   ** syntax documentation for more information on comment format.
   */
  string remove_comments(const string& s) const throw(FileParserException);


  /** Removes white spaces from a string.
   ** @param s The string to remove comments from.
   ** @return A string containing the same text as s but without whitespace characters.
   ** @exception FileParserException
   */
  string remove_white_space(const string& s) const throw(FileParserException);

  /** Processes environments variable specified by a "$" sign.
   ** @param s The string to preprocess.
   ** @return A string where all the environment variables have been substituted
   ** by their values.
   */
  string process_env_vars(const string& s) const throw(FileParserException);


  /** This method will test if a character at a certain position is inside double quotes.
   ** @param index The index (0 based) of the character to test.
   ** @param pair_list A list of int pairs (vector) representing the positions of the double quotes characters.
   ** @return true if a character at position index is inside double quotes, else false.
   ** @exception FileParserException
   ** @see get_string_brackets_positions
   */
  bool in_string(string::size_type index, list< vector<int> >& pair_list) const throw(FileParserException);


  /** This method is used to find the double quotes 'pairs' in a string.
   ** @param s A string in which to find double quotes pairs.
   ** @return A list of 2 int elements vectors, each vector representing the position of 2 double quotes characters.
   ** @exception FileParserException
   */
  list< vector<int> > get_string_brackets_positions(const string& s) const throw(FileParserException);


  /** This method will check if the alternance "<" ... ">" ... "<" ... ">" ... is repected.
   ** @param s The string to check.
   ** @return true if alternance is ok, else false
   ** @throw FileParserException
   */
  bool node_blocks_brackets_ok(const string& s) throw(FileParserException);


  /** This method is used to construct an indexed structure containing all the information
   ** needed to build a data-flow network from a string.
   ** @param file_content A string containing the preprocessed content of a data-flow network text file.
   ** @return A indexed container object containing all the information about the file's contents.  The keys of
   ** this map are the names of the networks and the values are FP_network_info_container objects.
   ** @exception FileParserException
   ** @see FP_network_info_container
   */
  map<string,FP_network_info_container> process(const string& file_content) throw(FileParserException);

  /** This method is used to construct and fill a param_info structure from
   ** a string.
   ** @param right_side The string that needs to be converted into a param_info structure.
   ** @param net_name The name of the network where the right_side was taken from.
   ** @param node_block The node block from where the right_side was taken from.
   ** @return A properly filled param_info structure.
   ** @throw FileParserException
   */
  param_info fill_param_info(const string& right_side, const string& net_name, const string& node_block) const throw(FileParserException);


  /** Use this method to check if a string is numeric and if it is, returns the type
   ** of the numeric value.
   ** @param s The string to inspect.
   ** @return 0 if non numeric, 1 if unsigned integer, 2 if signed integer, 3 if unsigned float, 4 if signed float
   */
  int type_of_num_string(const string& s) const;

 public:
  /*
  ** Constructors / Destructor
  */

  /// Default constructor
  FileParser(); 

  /// Copy constructor
  FileParser(const FileParser& fp);

  /// Destructor
  virtual ~FileParser();

  /*
  ** Accessors
  */

  /** Sets the verbose flag.
   **
   */
  void set_verbose(const bool b);

  /** Gets the value of the verbose flag.
   ** @return true if verbose mode is on, else returns false.
   */
  bool verbose() const;

  /** Gets the filename of the last text file that was parsed.
   ** @return The filename.
   */
  string getFilename() const;


  /** Gets the network information of the last text file that was parsed.
   ** @return An indexed container (map) object containing the information about the network.
   ** The keys of the map are the names of the networks and associated values are FP_network_info_container 
   ** structures containing information about each network.
   ** @memo Note that the parse method must first be called for this method to return information, else an
   ** empty map will be returned.
   */
  map<string,FP_network_info_container> get_networks_map() const;



  /*
  ** File Parsing stuff
  */

  /** This method is used to parse a data-flow network text file.
   ** @param filename The text file to parse.
   ** @return The number of networks found in the file.
   ** @exception FileParserException.
   ** @memo After this method is called, use the get_networks_map method to get the information.
   */
  int parse(const string& filename) throw(FileParserException);
};

#endif
